//  Parallel UPC version
//  Numerical Solution of heat conduction equation in
//  two dimensions
#include <stdio.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <upc.h>
#include <upc_collective.h>

#define min(A,B) ((A) < (B) ? (A) : (B))
#define max(A,B) ((A) > (B) ? (A) : (B))

#define KMAX 11
//#define KMAX 100

const int i_print = 0;

//const int imax = 100;
//const int itmax = 100000;
const int imax = 20;
const int itmax = 20000;
const int kmax = KMAX;

shared double *dphimax;
shared double dphimax_loc[THREADS];
shared [] double *shared phin[THREADS];

double dwalltime();
void print_field(double **phi, int imax, int kmax);

int main()
{
  double eps = 1.0e-08;
  double **phi;
  double *phin_loc;
// the above must be allocated so they can be used as
// double phi[imax_loc+1][kmax+1], phin[MYTHREAD][imax_loc][kmax];
  double dx,dy,dx2,dy2,dx2i,dy2i,dt,dphi;
  double mflops, ti;
  int i,k,it;

  int imax_loc;
/* parallel housekeeping */
  imax_loc = imax / THREADS;
  if (imax%THREADS > 0) imax_loc = imax_loc + 1;
  phi = malloc((imax_loc+1)*sizeof(void *)); 
  phin[MYTHREAD] = (shared [] double *) 
                    upc_alloc(imax_loc*kmax*sizeof(double)); 
  for (i=0; i<=imax_loc; i++) phi[i] = malloc((kmax+1)*sizeof(double));
  dphimax = (shared double *) upc_all_alloc(1, sizeof(double));
/*  */
  dx=1.0/kmax;
  dy=1.0/(1 + (imax_loc - 1) * THREADS);
  dx2=dx*dx;
  dy2=dy*dy;
  dx2i=1.0/dx2;
  dy2i=1.0/dy2;
  dt=min(dx2,dy2)/4.0;
/* start values 0.d0 */
  for (k=0;k<kmax;k++)
  {
    for (i=0;i<=imax_loc;i++)
    {
      phi[i][k]=0.0;
    }
  }
/* start values 1.d0 */
  for (i=0;i<=imax_loc;i++)
  {
    phi[i][kmax]=1.0;
  }
/* start values dx */
  if (MYTHREAD == 0) {
    phi[0][0]=0.0;
    for (k=1;k<kmax;k++) 
    {
      phi[0][k]=phi[0][k-1]+dx;
    }
  }
  if (MYTHREAD == THREADS - 1) {
    phi[imax_loc][0]=0.0;
    for (k=1;k<kmax;k++) 
    {
      phi[imax_loc][k]=phi[imax_loc][k-1]+dx;
    }
  }

  if (MYTHREAD == 0) {
    printf("\nHeat Conduction 2d\n");
    printf("\ndx = %12.4g, dy = %12.4g, dt = %12.4g, eps = %12.4g\n",
         dx,dy,dt,eps);
  }

  if (imax < 21 && kmax < 21 && MYTHREAD == i_print) {
    print_field(phi, imax_loc, kmax);
  }

  ti=dwalltime();

/* iteration */
  for (it=1;it<=itmax;it++)
  {
    dphimax_loc[MYTHREAD]=0.;
    phin_loc = (double *) phin[MYTHREAD];
    for (i=1;i<imax_loc;i++)
    {
      for (k=1;k<kmax;k++)
      {
        dphi=(phi[i+1][k]+phi[i-1][k]-2.*phi[i][k])*dy2i
            +(phi[i][k+1]+phi[i][k-1]-2.*phi[i][k])*dx2i;
        dphi=dphi*dt;
        dphimax_loc[MYTHREAD]=max(dphimax_loc[MYTHREAD],dphi);
//        *(phin[MYTHREAD]+kmax*i+k)=phi[i][k]+dphi;
        *(phin_loc+kmax*i+k)=phi[i][k]+dphi;
      }
    }
    upc_barrier;
    if (MYTHREAD > 0) upc_memget(&phi[0][0],
        phin[MYTHREAD-1]+kmax*(imax_loc-1),KMAX*sizeof(double)); 
    if (MYTHREAD < THREADS - 1) upc_memget(&phi[imax_loc][0],
        phin[MYTHREAD+1]+kmax,KMAX*sizeof(double)); 
/* save values */
    for (i=1;i<imax_loc;i++)
    {
      for (k=1;k<kmax;k++)
      {
//      phi[i][k]=*(phin[MYTHREAD]+kmax*i+k);
        phi[i][k]=*(phin_loc+kmax*i+k);
      }
    }

//    if (it%250 == 1) print_field(phi, imax_loc, kmax);
    upc_all_reduceD(dphimax,dphimax_loc, 
                    UPC_MAX,THREADS,1,NULL,UPC_IN_ALLSYNC | UPC_OUT_ALLSYNC);
    upc_barrier;
    if(*dphimax<eps) break;
  }

  ti=dwalltime() - ti;
  if (imax < 21 && kmax < 21 && MYTHREAD == i_print) {
    print_field(phi, imax_loc, kmax);
  }
  if (MYTHREAD == 0) {
    printf("\n%d iterations\n",it);
    printf("\nCPU time = %#12.4g sec\n",ti);
    mflops = 11.0 * (float) (kmax-1)*(float) (imax-1)*(float)it / 1.e6;
    printf("\nMFlop/s = %#12.4g \n",mflops/ti);
  }

  for (i=0; i<=imax_loc; i++) free(phi[i]);
  free (phi); 
  upc_free (phin[MYTHREAD]); 
  if (MYTHREAD == 0) upc_free (dphimax); 
}

double dwalltime() {
  struct rusage r;
  getrusage( RUSAGE_SELF, &r );
  return r.ru_utime.tv_sec + r.ru_utime.tv_usec*1.0e-6;
}

void print_field(double **phi, int imax, int kmax) {
  int i, k, kk, kl, kkk, l;
  kl=6;
  kkk=kl-1;
  for (k=0; k<=kmax;k+=kl) {
    if(k+kkk > kmax) kkk=kmax-k;
    printf("columns %5i to %5i\n",k,k+kkk);
    for (i=0; i<=imax;i++) {
      printf("%5i ",i);
      for (l=0; l<kl; l++) {
        printf("%12.4g", phi[i][k+l]);
      }
      printf("\n");
    }
  }
}
